# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

import bz2
import contextlib
import json

import fakeredis
import mock
from google.cloud import storage

from taar.logs.interfaces import ITAARCache
from taar.recommenders.locale_recommender import LocaleRecommender
from taar.recommenders.redis_cache import TAARCacheRedis
from taar.settings import TAAR_LOCALE_KEY, TAAR_LOCALE_BUCKET
from .noop_fixtures import (
    noop_taarcollab_dataload,
    noop_taarlite_dataload,
    noop_taarsimilarity_dataload,
    noop_taarensemble_dataload,
)

FAKE_LOCALE_DATA = {
    "te-ST": [
        ["{1e6b8bce-7dc8-481c-9f19-123e41332b72}", 0.1],
        ["some-other@nice-addon.com", 0.2],
        ["{66d1eed2-a390-47cd-8215-016e9fa9cc55}", 0.3],
        ["{5f1594c3-0d4c-49dd-9182-4fbbb25131a7}", 0.4],
    ],
    "en": [["other-addon@some-id.it", 0.3], ["some-uuid@test-addon.com", 0.1]],
}


def install_mock_data(ctx):
    ctx = ctx.child()

    byte_data = json.dumps(FAKE_LOCALE_DATA).encode("utf8")
    byte_data = bz2.compress(byte_data)

    client = storage.Client()
    bucket = client.get_bucket(TAAR_LOCALE_BUCKET)
    blob = bucket.blob(TAAR_LOCALE_KEY)
    blob.upload_from_string(byte_data)

    return ctx


@contextlib.contextmanager
def mock_locale_data(ctx):
    with contextlib.ExitStack() as stack:
        TAARCacheRedis._instance = None
        stack.enter_context(
            mock.patch.object(
                TAARCacheRedis, "_fetch_locale_data", return_value=FAKE_LOCALE_DATA,
            )
        )

        stack = noop_taarlite_dataload(stack)
        stack = noop_taarcollab_dataload(stack)
        stack = noop_taarsimilarity_dataload(stack)
        stack = noop_taarensemble_dataload(stack)

        # Patch fakeredis in
        stack.enter_context(
            mock.patch.object(
                TAARCacheRedis,
                "init_redis_connections",
                return_value={
                    0: fakeredis.FakeStrictRedis(db=0),
                    1: fakeredis.FakeStrictRedis(db=1),
                    2: fakeredis.FakeStrictRedis(db=2),
                },
            )
        )

        # Initialize redis
        cache = TAARCacheRedis.get_instance(ctx)
        cache.safe_load_data()
        ctx[ITAARCache] = cache
        yield stack


def test_can_recommend(test_ctx):
    with mock_locale_data(test_ctx):
        r = LocaleRecommender(test_ctx)

        # Test that we can't recommend if we have not enough client info.
        assert not r.can_recommend({})
        assert not r.can_recommend({"locale": []})

        # Check that we can recommend if the user has at least an addon.
        assert r.can_recommend({"locale": "en"})


def test_can_recommend_no_model(test_ctx):
    with mock_locale_data(test_ctx):
        r = LocaleRecommender(test_ctx)

        # We should never be able to recommend if something went
        # wrong with the model.
        assert not r.can_recommend({})
        assert not r.can_recommend({"locale": []})
        assert not r.can_recommend({"locale": "it"})


def test_recommendations(test_ctx):
    """Test that the locale recommender returns the correct
    locale dependent addons.

    The JSON output for this recommender should be a list of 2-tuples
    of (GUID, weight).
    """
    with mock_locale_data(test_ctx):
        r = LocaleRecommender(test_ctx)

        recommendations = r.recommend({"locale": "en"}, 10)

        # Make sure the structure of the recommendations is correct and that we
        # recommended the the right addon.
        assert isinstance(recommendations, list)
        assert len(recommendations) == len(FAKE_LOCALE_DATA["en"])

        # Make sure that the reported addons are the one from the fake data.
        for (addon_id, weight), (expected_id, expected_weight) in zip(
                recommendations, FAKE_LOCALE_DATA["en"]
        ):
            assert addon_id == expected_id
            assert weight == expected_weight


def test_recommender_extra_data(test_ctx):
    # Test that the recommender uses locale data from the "extra"
    # section if available.
    def validate_recommendations(data, expected_locale):
        # Make sure the structure of the recommendations is correct and that we
        # recommended the the right addon.
        data = sorted(data, key=lambda x: x[1], reverse=True)
        assert isinstance(data, list)
        assert len(data) == len(FAKE_LOCALE_DATA[expected_locale])

        # Make sure that the reported addons are the one from the fake data.
        for (addon_id, weight), (expected_id, expected_weight) in zip(
                data, FAKE_LOCALE_DATA[expected_locale]
        ):
            assert addon_id == expected_id
            assert weight == expected_weight

    with mock_locale_data(test_ctx):
        r = LocaleRecommender(test_ctx)
        recommendations = r.recommend({}, 10, extra_data={"locale": "en"})
        validate_recommendations(recommendations, "en")

        # Make sure that we favour client data over the extra data.
        recommendations = r.recommend(
            {"locale": "en"}, 10, extra_data={"locale": "te-ST"}
        )
        validate_recommendations(recommendations, "en")



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http://mozilla.org/MPL/2.0/.

import json
import six
import logging

import numpy as np
import scipy.stats

from taar.logs.interfaces import ITAARCache
from taar.recommenders.similarity_recommender import (
    CATEGORICAL_FEATURES,
    CONTINUOUS_FEATURES,
    SimilarityRecommender,
)

from .similarity_data import CONTINUOUS_FEATURE_FIXTURE_DATA
from .similarity_data import CATEGORICAL_FEATURE_FIXTURE_DATA

import fakeredis
import mock
import contextlib
from .noop_fixtures import (
    noop_taarcollab_dataload,
    noop_taarlite_dataload,
    noop_taarlocale_dataload,
    noop_taarensemble_dataload,
)
from taar.recommenders.redis_cache import TAARCacheRedis


def noop_loaders(stack):
    stack = noop_taarlocale_dataload(stack)
    stack = noop_taarcollab_dataload(stack)
    stack = noop_taarensemble_dataload(stack)
    stack = noop_taarlite_dataload(stack)
    return stack


def generate_fake_lr_curves(num_elements, ceiling=10.0):
    """
    Generate a mock likelihood ratio (LR) curve that can be used for
    testing.
    """
    lr_index = list(np.linspace(0, ceiling, num_elements))

    # This sets up a normal distribution with a mean of 0.5 and std
    # deviation of 0.5
    numerator_density = [scipy.stats.norm.pdf(float(i), 0.5, 0.5) for i in lr_index]

    # This sets up a normal distribution with a mean of 5.0 and std
    # deviation of 1.5.  So this is right shifted and has a wide std
    # deviation compared to the first curve.

    # This results in a small overlap.  Using ranges between 0.5 to
    # 5.0 will yield large values to small values.
    denominator_density = [scipy.stats.norm.pdf(float(i), 5.0, 1.5) for i in lr_index]
    return list(zip(lr_index, zip(numerator_density, denominator_density)))


def generate_a_fake_taar_client():
    return {
        "client_id": "test-client-001",
        "activeAddons": [],
        "geo_city": "brasilia-br",
        "subsession_length": 4911,
        "locale": "br-PT",
        "os": "mac",
        "bookmark_count": 7,
        "tab_open_count": 4,
        "total_uri": 222,
        "unique_tlds": 21,
    }


@contextlib.contextmanager
def mock_install_no_data(ctx):
    with contextlib.ExitStack() as stack:
        TAARCacheRedis._instance = None
        stack.enter_context(
            mock.patch.object(TAARCacheRedis, "_fetch_similarity_donors", return_value="", )
        )

        stack.enter_context(
            mock.patch.object(TAARCacheRedis, "_fetch_similarity_lrcurves", return_value="", )
        )

        stack = noop_loaders(stack)

        # Patch fakeredis in
        stack.enter_context(
            mock.patch.object(
                TAARCacheRedis,
                "init_redis_connections",
                return_value={
                    0: fakeredis.FakeStrictRedis(db=0),
                    1: fakeredis.FakeStrictRedis(db=1),
                    2: fakeredis.FakeStrictRedis(db=2),
                },
            )
        )

        # Initialize redis
        cache = TAARCacheRedis.get_instance(ctx)
        cache.safe_load_data()
        ctx[ITAARCache] = cache
        yield stack


@contextlib.contextmanager
def mock_install_categorical_data(ctx):
    with contextlib.ExitStack() as stack:
        TAARCacheRedis._instance = None
        stack.enter_context(
            mock.patch.object(
                TAARCacheRedis,
                "_fetch_similarity_donors",
                return_value=CATEGORICAL_FEATURE_FIXTURE_DATA,
            )
        )

        stack.enter_context(
            mock.patch.object(
                TAARCacheRedis,
                "_fetch_similarity_lrcurves",
                return_value=generate_fake_lr_curves(1000),
            )
        )
        stack = noop_loaders(stack)

        # Patch fakeredis in
        stack.enter_context(
            mock.patch.object(
                TAARCacheRedis,
                "init_redis_connections",
                return_value={
                    0: fakeredis.FakeStrictRedis(db=0),
                    1: fakeredis.FakeStrictRedis(db=1),
                    2: fakeredis.FakeStrictRedis(db=2),
                },
            )
        )

        # Initialize redis
        cache = TAARCacheRedis.get_instance(ctx)
        cache.safe_load_data()
        ctx[ITAARCache] = cache
        yield stack


@contextlib.contextmanager
def mock_install_continuous_data(ctx):
    cts_data = CONTINUOUS_FEATURE_FIXTURE_DATA
    lrs_data = generate_fake_lr_curves(1000)

    with contextlib.ExitStack() as stack:
        TAARCacheRedis._instance = None
        stack.enter_context(
            mock.patch.object(
                TAARCacheRedis, "_fetch_similarity_donors", return_value=cts_data,
            )
        )

        stack.enter_context(
            mock.patch.object(
                TAARCacheRedis, "_fetch_similarity_lrcurves", return_value=lrs_data,
            )
        )
        stack = noop_loaders(stack)

        # Patch fakeredis in
        stack.enter_context(
            mock.patch.object(
                TAARCacheRedis,
                "init_redis_connections",
                return_value={
                    0: fakeredis.FakeStrictRedis(db=0),
                    1: fakeredis.FakeStrictRedis(db=1),
                    2: fakeredis.FakeStrictRedis(db=2),
                },
            )
        )

        # Initialize redis
        cache = TAARCacheRedis.get_instance(ctx)
        cache.safe_load_data()
        ctx[ITAARCache] = cache
        yield stack


def test_soft_fail(test_ctx, caplog):
    # Create a new instance of a SimilarityRecommender.
    with mock_install_no_data(test_ctx):
        r = SimilarityRecommender(test_ctx)

        # Don't recommend if the source files cannot be found.
        assert not r.can_recommend({})


def test_can_recommend(test_ctx, caplog):
    caplog.set_level(logging.INFO)

    # Create a new instance of a SimilarityRecommender.
    with mock_install_continuous_data(test_ctx):
        r = SimilarityRecommender(test_ctx)

        # Test that we can't recommend if we have not enough client info.
        assert not r.can_recommend({})

        # Test that we can recommend for a normal client.
        assert r.can_recommend(generate_a_fake_taar_client())

        # Check that we can not recommend if any required client field is missing.
        required_fields = CATEGORICAL_FEATURES + CONTINUOUS_FEATURES

        for required_field in required_fields:
            profile_without_x = generate_a_fake_taar_client()

            # Make an empty value in a required field in the client info dict.
            profile_without_x[required_field] = None
            assert not r.can_recommend(profile_without_x)

            # Completely remove (in place) the entire required field from the dict.
            del profile_without_x[required_field]
            assert not r.can_recommend(profile_without_x)


def test_recommendations(test_ctx):
    # Create a new instance of a SimilarityRecommender.
    with mock_install_continuous_data(test_ctx):
        r = SimilarityRecommender(test_ctx)

        recommendation_list = r.recommend(generate_a_fake_taar_client(), 1)

        assert isinstance(recommendation_list, list)
        assert len(recommendation_list) == 1

        recommendation, weight = recommendation_list[0]

        # Make sure that the reported addons are the expected ones from the most similar donor.
        assert "{test-guid-1}" == recommendation
        assert type(weight) == np.float64


def test_get_lr(test_ctx):
    # Tests that the likelihood ratio values are not empty for extreme values and are realistic.
    with mock_install_continuous_data(test_ctx):
        r = SimilarityRecommender(test_ctx)
        cache = r._get_cache({})
        assert r.get_lr(0.0001, cache) is not None
        assert r.get_lr(10.0, cache) is not None
        assert r.get_lr(0.001, cache) > r.get_lr(5.0, cache)


def test_compute_clients_dist(test_ctx):
    # Test the distance function computation.
    with mock_install_continuous_data(test_ctx):
        r = SimilarityRecommender(test_ctx)
        test_clients = [
            {
                "client_id": "test-client-002",
                "activeAddons": [],
                "geo_city": "sfo-us",
                "subsession_length": 1,
                "locale": "en-US",
                "os": "windows",
                "bookmark_count": 1,
                "tab_open_count": 1,
                "total_uri": 1,
                "unique_tlds": 1,
            },
            {
                "client_id": "test-client-003",
                "activeAddons": [],
                "geo_city": "brasilia-br",
                "subsession_length": 1,
                "locale": "br-PT",
                "os": "windows",
                "bookmark_count": 10,
                "tab_open_count": 1,
                "total_uri": 1,
                "unique_tlds": 1,
            },
            {
                "client_id": "test-client-004",
                "activeAddons": [],
                "geo_city": "brasilia-br",
                "subsession_length": 100,
                "locale": "br-PT",
                "os": "windows",
                "bookmark_count": 10,
                "tab_open_count": 10,
                "total_uri": 100,
                "unique_tlds": 10,
            },
        ]
        per_client_test = []

        # Compute a different set of distances for each set of clients.
        cache = r._get_cache({})
        for tc in test_clients:
            test_distances = r.compute_clients_dist(tc, cache)
            assert len(test_distances) == len(CONTINUOUS_FEATURE_FIXTURE_DATA)
            per_client_test.append(test_distances[2][0])

        # Ensure the different clients also had different distances to a specific donor.
        assert per_client_test[0] >= per_client_test[1] >= per_client_test[2]


def test_distance_functions(test_ctx):
    # Tests the similarity functions via expected output when passing
    # modified client data.
    with mock_install_continuous_data(test_ctx):
        r = SimilarityRecommender(test_ctx)

        # Generate a fake client.
        test_client = generate_a_fake_taar_client()
        recs = r.recommend(test_client, 10)
        assert len(recs) > 0

        # Make it a generally poor match for the donors.
        test_client.update(
            {"total_uri": 10, "bookmark_count": 2, "subsession_length": 10}
        )

        all_client_values_zero = test_client
        # Make all categorical variables non-matching with any donor.
        all_client_values_zero.update(
            {key: "zero" for key in test_client.keys() if key in CATEGORICAL_FEATURES}
        )
        recs = r.recommend(all_client_values_zero, 10)
        assert len(recs) == 0

        # Make all continuous variables equal to zero.
        all_client_values_zero.update(
            {key: 0 for key in test_client.keys() if key in CONTINUOUS_FEATURES}
        )
        recs = r.recommend(all_client_values_zero, 10)
        assert len(recs) == 0

        # Make all categorical variables non-matching with any donor.
        all_client_values_high = test_client
        all_client_values_high.update(
            {
                key: "one billion"
                for key in test_client.keys()
                if key in CATEGORICAL_FEATURES
            }
        )
        recs = r.recommend(all_client_values_high, 10)
        assert len(recs) == 0

        # Make all continuous variables equal to a very high numerical value.
        all_client_values_high.update(
            {key: 1e60 for key in test_client.keys() if key in CONTINUOUS_FEATURES}
        )
        recs = r.recommend(all_client_values_high, 10)
        assert len(recs) == 0

        # Test for 0.0 values if j_c is not normalized and j_d is fine.
        j_c = 0.0
        j_d = 0.42
        assert abs(j_c * j_d) == 0.0
        assert abs((j_c + 0.01) * j_d) != 0.0


def test_weights_continuous(test_ctx):
    # Create a new instance of a SimilarityRecommender.
    with mock_install_continuous_data(test_ctx):
        r = SimilarityRecommender(test_ctx)

        # In the ensemble method recommendations should be a sorted list of tuples
        # containing [(guid, weight), (guid, weight)... (guid, weight)].
        recommendation_list = r.recommend(generate_a_fake_taar_client(), 2)
        with open("/tmp/similarity_recommender.json", "w") as fout:
            fout.write(json.dumps(recommendation_list))

        # Make sure the structure of the recommendations is correct and
        # that we recommended the the right addons.

        assert len(recommendation_list) == 2
        for recommendation, weight in recommendation_list:
            assert isinstance(recommendation, six.string_types)
            assert isinstance(weight, float)

        # Test that sorting is appropriate.
        rec0 = recommendation_list[0]
        rec1 = recommendation_list[1]

        rec0_weight = rec0[1]
        rec1_weight = rec1[1]

        # Duplicate presence of test-guid-1 should mean rec0_weight is double
        # rec1_weight, and both should be greater than 1.0

        assert rec0_weight > rec1_weight > 1.0


def test_weights_categorical(test_ctx):
    """
    This should get :
        ["{test-guid-1}", "{test-guid-2}", "{test-guid-3}", "{test-guid-4}"],
        ["{test-guid-9}", "{test-guid-10}", "{test-guid-11}", "{test-guid-12}"]
    from the first two entries in the sample data where the geo_city
    data

    """
    # Create a new instance of a SimilarityRecommender.
    with mock_install_categorical_data(test_ctx):
        r = SimilarityRecommender(test_ctx)

        # In the ensemble method recommendations should be a sorted list of tuples
        # containing [(guid, weight), (guid, weight)... (guid, weight)].
        recommendation_list = r.recommend(generate_a_fake_taar_client(), 2)

        assert len(recommendation_list) == 2
        # Make sure the structure of the recommendations is correct and that we recommended the the right addons.
        for recommendation, weight in recommendation_list:
            assert isinstance(recommendation, six.string_types)
            assert isinstance(weight, float)

        # Test that sorting is appropriate.
        rec0 = recommendation_list[0]
        rec1 = recommendation_list[1]

        rec0_weight = rec0[1]
        rec1_weight = rec1[1]

        assert rec0_weight > rec1_weight > 0



from sys import exc_info

from taar.logs.interfaces import IMozLogging


class EmergencyLogger:
    """
    We need this one to get rid of python logging dependency in Ensemble spark job
     (see more detailed explanation in readme).
    It uses only print and logs only errors and warnings
    """
    def debug(self, msg, *args, **kwags):
        pass

    def info(self, msg, *args, **kwags):
        pass

    def warn(self, msg, *args, **kwags):
        print(f'WARN: {msg}')

    def warning(self, msg, *args, **kwags):
        self.warn(msg)

    def error(self, msg, e=None, *args, **kwags):
        print(f'ERROR: {msg}, {e or exc_info()}')

    def exception(self, msg, *args, **kwargs):
        self.error(msg, *args, **kwargs)


class LoggingStub(IMozLogging):
    def __init__(self, ctx):
        pass

    def get_logger(self, name):
        return EmergencyLogger()

    def set_log_level(self, level):
        pass



#!/usr/bin/env python
from taar.recommenders.cache import TAARCache
from taar.recommenders.redis_cache import TAARCacheRedis
from taar.context import default_context
import click


@click.command()
@click.option("--reset", is_flag=True, help="Reset the redis cache to an empty state")
@click.option("--load", is_flag=True, help="Load data into redis")
@click.option("--info", is_flag=True, help="Display information about the cache state")
def main(reset, load, info):
    """
    Manage the TAAR+TAARLite redis cache.

    This expecte that the following enviroment variables are set:

    REDIS_HOST
    REDIS_PORT
    """
    if not (reset or load or info):
        print("No options were set!")
        return

    ctx = default_context(TAARCacheRedis)
    cache = ctx[TAARCache]

    if reset:
        if cache.reset():
            print("Successfully flushed db0 bookkeeping database.")
        else:
            print("Error while flushign db0 bookkeeping database.")
    if load:
        cache.safe_load_data()
    if info:
        cache.info()


if __name__ == "__main__":
    main()



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

import itertools

import markus
from sentry_sdk import capture_exception

from taar.logs.interfaces import IMozLogging, ITAARCache
from taar.recommenders.debug import log_timer_debug
from taar.utils import hasher
from taar.recommenders.base_recommender import AbstractRecommender

metrics = markus.get_metrics("taar")


def is_test_client(client_id):
    """ any client_id where the GUID is composed of a single digit
    (repeating) is a test id """
    return len(set(client_id.replace("-", ""))) == 1


class EnsembleRecommender(AbstractRecommender):
    """
    The EnsembleRecommender is a collection of recommenders where the
    results from each recommendation is amplified or dampened by a
    factor.  The aggregate results are combines and used to recommend
    addons for users.
    """

    def __init__(self, ctx):
        self.RECOMMENDER_KEYS = ["collaborative", "similarity", "locale"]
        self._ctx = ctx

        self._redis_cache = self._ctx[ITAARCache]
        self.logger = self._ctx[IMozLogging].get_logger("taar.ensemble")

        assert "recommender_factory" in self._ctx

        self._init_from_ctx()

    def _get_cache(self, extra_data):
        tmp = extra_data.get("cache", None)
        if tmp is None:
            tmp = self._redis_cache.cache_context()
        return tmp

    def getWeights(self):
        return self._redis_cache.ensemble_weights()

    def _init_from_ctx(self):
        # Copy the map of the recommenders
        self._recommender_map = {}

        recommender_factory = self._ctx["recommender_factory"]
        for rkey in self.RECOMMENDER_KEYS:
            self._recommender_map[rkey] = recommender_factory.create(rkey)

        self.logger.info("EnsembleRecommender initialized")

    def can_recommend(self, client_data, extra_data={}):
        """The ensemble recommender is always going to be
        available if at least one recommender is available"""
        result = sum(
            [
                self._recommender_map[rkey].can_recommend(client_data)
                for rkey in self.RECOMMENDER_KEYS
            ]
        )
        self.logger.debug("Ensemble can_recommend: {}".format(result))
        return result

    @metrics.timer_decorator("ensemble_recommend")
    def recommend(self, client_data, limit, extra_data={}):
        cache = self._get_cache(extra_data)
        client_id = client_data.get("client_id", "no-client-id")

        if is_test_client(client_id):
            whitelist = cache["whitelist"]
            samples = whitelist[:limit]
            self.logger.info("Test ID detected [{}]".format(client_id))

            # Compute a stable weight for any whitelisted addon based
            # on the sha256 hash of the GUID
            p = [(int(hasher(s), 16) % 100) / 100.0 for s in samples]
            results = list(zip(samples, p))
        else:
            try:
                metrics.incr("error_ensemble", value=1)
                results = self._recommend(client_data, limit, extra_data)
            except Exception as e:
                results = []
                self.logger.exception(f"Ensemble recommender crashed for {client_id}")
                capture_exception(e)
        return results

    def _recommend(self, client_data, limit, extra_data={}):
        """
        Ensemble recommendations are aggregated from individual
        recommenders.  The ensemble recommender applies a weight to
        the recommendation outputs of each recommender to reorder the
        recommendations to be a better fit.

        The intuitive understanding is that the total space of
        recommended addons across all recommenders will include the
        'true' addons that should be recommended better than any
        individual recommender.  The ensemble method simply needs to
        weight each recommender appropriate so that the ordering is
        correct.
        """
        cache = self._get_cache(extra_data)
        self.logger.debug("Ensemble recommend invoked")
        preinstalled_addon_ids = client_data.get("installed_addons", [])

        # Compute an extended limit by adding the length of
        # the list of any preinstalled addons.
        extended_limit = limit + len(preinstalled_addon_ids)

        flattened_results = []
        ensemble_weights = cache["ensemble_weights"]

        for rkey in self.RECOMMENDER_KEYS:
            self._recommend_single(client_data, ensemble_weights, extended_limit, extra_data, flattened_results, rkey)

        # Sort the results by the GUID
        flattened_results.sort(key=lambda item: item[0])

        # group by the guid, sum up the weights for recurring GUID
        # suggestions across all recommenders
        guid_grouper = itertools.groupby(flattened_results, lambda item: item[0])

        ensemble_suggestions = []
        for (guid, guid_group) in guid_grouper:
            weight_sum = sum([v for (g, v) in guid_group])
            item = (guid, weight_sum)
            ensemble_suggestions.append(item)

        # Sort in reverse order (greatest weight to least)
        ensemble_suggestions.sort(key=lambda x: -x[1])

        filtered_ensemble_suggestions = [
            (guid, weight)
            for (guid, weight) in ensemble_suggestions
            if guid not in preinstalled_addon_ids
        ]

        results = filtered_ensemble_suggestions[:limit]

        log_data = (
            client_data["client_id"],
            extra_data.get("guid_randomization", False),
            str(ensemble_weights),
            str([r[0] for r in results]),
        )
        self.logger.debug(
            "client_id: [%s], guid_randomization: [%s], ensemble_weight: [%s], guids: [%s]"
            % log_data
        )
        return results

    def _recommend_single(self, client_data, ensemble_weights, extended_limit, extra_data, flattened_results, rkey):
        with log_timer_debug(f"{rkey} recommend invoked", self.logger):
            recommender = self._recommender_map[rkey]
            if not recommender.can_recommend(client_data, extra_data):
                return
            with metrics.timer(f"{rkey}_recommend"):
                try:
                    raw_results = recommender.recommend(
                        client_data, extended_limit, extra_data
                    )
                except Exception as e:
                    metrics.incr(f"error_{rkey}", value=1)
                    self.logger.exception(
                        "{} recommender crashed for {}".format(rkey,
                                                               client_data.get("client_id", "no-client-id")
                                                               ),
                        e,
                    )
            reweighted_results = []
            for guid, weight in raw_results:
                item = (guid, weight * ensemble_weights[rkey])
                reweighted_results.append(item)
            flattened_results.extend(reweighted_results)



from setuptools import find_packages, setup

setup(
    name="mozilla-taar3",
    use_scm_version=False,
    version="1.0.5",
    setup_requires=["setuptools_scm", "pytest-runner"],
    tests_require=["pytest"],
    include_package_data=True,
    packages=find_packages(exclude=["tests", "tests/*"]),
    description="Telemetry-Aware Addon Recommender",
    author="Mozilla Foundation",
    author_email="fx-data-dev@mozilla.org",
    url="https://github.com/mozilla/taar",
    license="MPL 2.0",
    install_requires=[],
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Environment :: Web Environment :: Mozilla",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: Mozilla Public License 2.0 (MPL 2.0)",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Topic :: Internet :: WWW/HTTP",
        "Topic :: Scientific/Engineering :: Information Analysis",
    ],
    entry_points="""
    [taarapi_app]
    app=taar.plugin:configure_plugin
    """,
    scripts=["bin/taar-redis.py"],
    zip_safe=False,
)
